from tooldelta.plugin_load.classic_plugin import Plugin
from tooldelta.color_print import Print
from tooldelta.plugin_load import plugin_is_enabled
import re
import os
import sys

NOT_IMPORTALL_RULE = re.compile(r"from .* import \*")


def _import_original_dotcs_plugin(
    plugin_code: str, old_dotcs_env: dict, module_env: dict, plugin_group
):
    # dotcs插件太不规范了!!
    plugin_body = Plugin()
    plugin_body.dotcs_old_type = True
    old_dotcs_env.update(module_env)
    runcode_tmp = {}
    plugin_code_lines = plugin_code.split("\n")
    _dotcs_runcode = {}
    evts = {}
    packetFuncs = []
    while 1:
        if "" in plugin_code_lines:
            plugin_code_lines.remove("")
        else:
            break
    plugin_start = -1
    plugin_end = 0
    plugin_prev_type = ""

    for line in range(len(plugin_code_lines)):
        if "import *" in plugin_code_lines[line]:
            Print.print_war(
                f"DotCS插件 存在异常的代码 'import *', 已自动替换为 pass: §7{line} §f|{plugin_code_lines[line]}"
            )
            plugin_code_lines[line] = NOT_IMPORTALL_RULE.sub(
                "pass", plugin_code_lines[line]
            )
        if plugin_code_lines[line].startswith("# PLUGIN TYPE: "):
            if plugin_start + 1:
                plugin_end = line
                runcode_tmp[plugin_prev_type] = "\n".join(
                    plugin_code_lines[plugin_start:plugin_end]
                )
            plugin_start = line + 1
            plugin_prev_type = plugin_code_lines[line].strip("# PLUGIN TYPE: ")
    if plugin_start + 1:
        plugin_end = len(plugin_code_lines)
        runcode_tmp[plugin_prev_type] = "\n".join(
            plugin_code_lines[plugin_start:plugin_end]
        )

    old_dotcs_env.update({"plugin_group": plugin_group})
    fun_exec_code = ""
    for k in runcode_tmp:
        match k:
            case "def":
                fun_exec_code = "def on_def():\n "
            case "init":
                fun_exec_code = "def on_inject():\n "
            case "player prejoin":
                fun_exec_code = "def on_player_prejoin(playername):\n "
            case "player join":
                fun_exec_code = "def on_player_join(playername):\n "
            case "player message":
                fun_exec_code = "def on_player_message(playername, msg):\n "
            case "player leave":
                fun_exec_code = "def on_player_leave(playername):\n "
            case "player death":
                fun_exec_code = "def on_player_death(playername, killer, msg):\n "
            case "repeat 1s":
                fun_exec_code = "def repeat1s():\n "
            case "repeat 10s":
                fun_exec_code = "def repeat10s():\n "
            case "repeat 30s":
                fun_exec_code = "def repeat30s():\n "
            case "repeat 1m":
                fun_exec_code = "def repeat1m():\n "
            case _:
                if k.startswith("packet"):
                    try:
                        pktID = int(k.split()[1])
                        if pktID == -1:
                            Print.print_war("§c无法监听任意数据包, 已跳过")
                        else:
                            fun_exec_code = (
                                f"def packet_{pktID}(jsonPkt):\n packetType={pktID}\n "
                            )
                            plugin_body.add_req_listen_packet(pktID)
                            packetFuncs.append((pktID, f"packet_{pktID}"))
                    except:
                        Print.print_war(f"§c不合法的监听数据包ID： {k}, 已跳过")
                else:
                    raise Exception(f"无法识别的DotCS插件事件样式： {k}")
        # DotCS 有太多的插件都会出现作用域问题, 在此只能这么修复了
        p_code = (
            fun_exec_code
            + """globals().update(plugin_group.dotcs_global_vars)\n """
            + runcode_tmp[k].replace("\n", "\n ")
            + "\n plugin_group.dotcs_global_vars.update(locals())"
        )
        try:
            exec(p_code, old_dotcs_env, _dotcs_runcode)
        except Exception as err:
            Print.print_err(f"DotCS插件 <{plugin_body.name}> 出错: {err}")
            raise
    newPacketFuncs = []
    for pkt, funcname in packetFuncs:
        newPacketFuncs.append((pkt, _dotcs_runcode[funcname]))
    for codetype in [
        "on_def",
        "on_inject",
        "on_player_prejoin",
        "on_player_join",
        "on_player_message",
        "on_player_death",
        "on_player_leave",
    ]:
        if _dotcs_runcode.get(codetype):
            evts[codetype] = [plugin_body.name, _dotcs_runcode[codetype]]
        if _dotcs_runcode.get("repeat1s"):
            evts["repeat1s"] = [plugin_body.name, _dotcs_runcode["repeat1s"]]
        if _dotcs_runcode.get("repeat10s"):
            evts["repeat10s"] = [plugin_body.name, _dotcs_runcode["repeat10s"]]
        if _dotcs_runcode.get("repeat30s"):
            evts["repeat30s"] = [plugin_body.name, _dotcs_runcode["repeat30s"]]
        if _dotcs_runcode.get("repeat1m"):
            evts["repeat1m"] = [plugin_body.name, _dotcs_runcode["repeat1m"]]
    return plugin_body, evts, newPacketFuncs


def read_plugin_from_old(plugin_grp, module_env: dict):
    PLUGIN_PATH = os.path.join(os.getcwd(), "插件文件/原DotCS插件")
    sys.path.append(PLUGIN_PATH)
    dotcs_env = plugin_grp.linked_frame._get_old_dotcs_env()
    files = os.listdir(PLUGIN_PATH)
    files.sort()
    for file in files:
        if not plugin_is_enabled(file):
            continue
        try:
            if file.endswith(".py"):
                Print.print_inf(f"§6载入DotCS插件: {file.strip('.py')}", end="\r")
                with open(os.path.join(PLUGIN_PATH, file), "r", encoding="utf-8") as f:
                    code = f.read()
                plugin, evts, pkfuncs = _import_original_dotcs_plugin(
                    code, dotcs_env, module_env, plugin_grp
                )
                plugin.name = file.strip(".py")
                if (
                    "repeat10s" in evts
                    or "repeat1s" in evts
                    or "repeat30s" in evts
                    or "repeat1m" in evts
                ):
                    evtnew = evts.copy()
                    for i in evtnew:
                        if i.startswith("repeat"):
                            del evts[i]
                            plugin_grp.dotcs_repeat_threadings[
                                i.strip("repeat")
                            ].append(evtnew[i])
                for pkt, func in pkfuncs:
                    plugin_grp.add_listen_packet_id(pkt)
                    plugin_grp.add_listen_packet_func(pkt, func)
                    Print.print_suc(
                        f"[DotCS插件 特殊数据包监听] 添加成功: {plugin.name} <- {func.__name__}"
                    )
                for k, v in evts.items():
                    plugin_grp.plugins_funcs[k].append(v)
                plugin_grp.add_plugin(plugin)
                plugin_grp.dotcs_plugin_loaded_num += 1
                Print.print_suc(f"§a成功载入插件 §2<DotCS> §a{plugin.name}")
        except Exception as err:
            try:
                plugin_name = plugin.name  # type: ignore
            except:
                plugin_name = file
            Print.print_err(f"§c加载插件 {plugin_name} 出现问题: {err}")
            raise



import requests
import json
import os
import platform
import shutil
import tempfile
from tooldelta import urlmethod
from tooldelta.builtins import Builtins
from tooldelta.color_print import Print
from tooldelta.plugin_load import PluginRegData
from tooldelta.cfg import Cfg
from typing import Dict

if platform.system().lower() == "windows":
    CLS_CMD = "cls"
else:
    CLS_CMD = "clear"

clear_screen = lambda: os.system(CLS_CMD)


def _path_dir(path: str):
    if "/" not in path:
        return None
    return "/".join(path.split("/")[:-1])


def _url_join(*urls):
    return "/".join(urls)


def _get_json_from_url(url: str):
    try:
        resp = requests.get(url).text
    except requests.RequestException:
        raise Exception("URL请求失败")
    try:
        return json.loads(resp)
    except json.JSONDecodeError:
        raise Exception(f"服务器返回了不正确的答复: {resp}")


class PluginMarket:
    def enter_plugin_market(self, source_url: str = None):
        Print.print_inf("正在连接到插件市场..")
        try:
            market_datas = self.get_datas_from_market(source_url)
            plugins_list = list(market_datas["MarketPlugins"].items())
            all_indexes = len(plugins_list)
            now_index = 0
            sum_pages = int((all_indexes - 1) / 8) + 1
            now_page = 0
            last_operation = ""
            while True:
                clear_screen()
                Print.print_inf(
                    market_datas["SourceName"] + ": " + market_datas["Greetings"]
                )
                now_page = int(now_index / 8) + 1
                for i in range(now_index, now_index + 8):
                    if i in range(all_indexes):
                        plugin_data = PluginRegData(
                            plugins_list[i][0], plugins_list[i][1]
                        )
                        Print.print_inf(
                            f" {i + 1}. §e{plugin_data.name} §av{plugin_data.version_str} §b@{plugin_data.author} §d{plugin_data.plugin_type_str}插件",
                            need_log=False,
                        )
                    else:
                        Print.print_inf("")
                Print.print_inf(f"§f第 {now_page} / {sum_pages} 页, 输入 §b+§f/§b- §f翻页")
                Print.print_inf("§f输入插件序号选择插件, 以查看详情", need_log=False)
                last_operation = (
                    (
                        input(Print.fmt_info("回车键继续上次操作, §bq§f 退出, 请输入:", "§f 输入 "))
                        or last_operation
                    )
                    .lower()
                    .strip()
                )
                if last_operation == "+":
                    now_index += 8
                elif last_operation == "-":
                    now_index -= 8
                elif last_operation == "q":
                    break
                else:
                    res = Builtins.try_int(last_operation)
                    if res:
                        if res in range(1, all_indexes + 1):
                            r = self.choice_plugin(
                                PluginRegData(
                                    plugins_list[res - 1][0], plugins_list[res - 1][1]
                                ),
                                market_datas["MarketPlugins"],
                            )
                            if r:
                                Print.print_inf("下载插件后重启ToolDelta才能生效", need_log=False)
                                r = input(Print.fmt_info("§f输入 §cq §f退出, 其他则返回插件市场"))
                                if r.lower() == "q":
                                    break
                        else:
                            Print.print_err("超出序号范围")
                if now_index >= all_indexes:
                    now_index = 0
                elif now_index < 0:
                    now_index = max(all_indexes - 8, 0)
        except KeyError as err:
            Print.print_err(f"获取插件市场插件出现问题: 键值对错误: {err}")
            return
        except Exception as err:
            Print.print_err(f"获取插件市场插件出现问题: {err}")
            return
        clear_screen()
        Print.print_suc("已从插件市场返回 ToolDelta 控制台.")

    def get_datas_from_market(self, source_url: str = None):
        if source_url is None:
            source_url = Cfg().get_cfg("ToolDelta基本配置.json", {"插件市场源": str})["插件市场源"]
        market_datas = _get_json_from_url(
            _url_join(source_url, "market_tree.json")
        )
        self.plugins_download_url = market_datas["DownloadRefURL"]
        return market_datas

    def choice_plugin(self, plugin_data: PluginRegData, all_plugins_dict: dict):
        pre_plugins_str = (
            ", ".join([f"{k}§7v{v}" for k, v in plugin_data.pre_plugins.items()]) or "无"
        )
        clear_screen()
        Print.print_inf(
            f"{plugin_data.name} v{plugin_data.version_str}", need_log=False
        )
        Print.print_inf(
            f"作者: §f{plugin_data.author}§7, 版本: §f{plugin_data.version_str} §b{plugin_data.plugin_type_str}",
            need_log=False,
        )
        Print.print_inf(f"前置插件: §f{pre_plugins_str}", need_log=False)
        Print.print_inf(f"介绍: {plugin_data.description}", need_log=False)
        Print.print_inf("", need_log=False)
        res = input(Print.fmt_info("§f下载 = §aY§f, 取消 = §cN§f, 请输入:")).lower().strip()
        if res == "y":
            self.download_plugin(plugin_data, all_plugins_dict)
            return True
        return False

    def download_plugin(
        self,
        plugin_data: PluginRegData,
        all_plugins_dict: Dict[str, str],
    ):
        download_paths = self.find_dirs(plugin_data)
        for plugin_name, _ in plugin_data.pre_plugins.items():
            Print.print_inf(f"正在下载 {plugin_data.name} 的前置插件 {plugin_name}")
            self.download_plugin(
                PluginRegData(plugin_name, all_plugins_dict[plugin_name]),
                all_plugins_dict,
            )
        cache_dir = tempfile.mkdtemp()
        try:
            for paths in download_paths:
                if not paths.strip():
                    # 不可能出现的状况, 出现了证明是你的问题
                    Print.print_war("下载路径为空, 跳过")
                    continue
                url = _url_join(self.plugins_download_url, paths)
                # Determine download path based on plugin type
                match plugin_data.plugin_type:
                    case "classic":
                        download_path = os.path.join(
                            os.getcwd(), "插件文件", "ToolDelta组合式插件"
                        )
                    case "dotcs":
                        download_path = os.path.join(os.getcwd(), "插件文件", "原DotCS插件")
                    case "injected":
                        download_path = os.path.join(
                            os.getcwd(), "插件文件", "ToolDelta注入式插件"
                        )
                    case _:
                        raise Exception(
                            f"未知插件类型: {plugin_data.plugin_type}, 你可能需要通知ToolDelta项目开发组解决"
                        )
                os.makedirs(os.path.join(cache_dir, plugin_data.name), exist_ok=True)
                path_last = _path_dir(paths)
                if path_last is not None:
                    # 自动创建文件夹
                    folder_path = os.path.join(cache_dir, path_last)
                    os.makedirs(folder_path, exist_ok=True)
                urlmethod.download_unknown_file(url, os.path.join(cache_dir, paths))
            # Move downloaded files to target download path
            target_path = download_path
            os.makedirs(target_path, exist_ok=True)
            for root, _, files in os.walk(cache_dir):
                for filename in files:
                    source_file = os.path.join(root, filename)
                    target_file = os.path.join(
                        target_path, os.path.relpath(source_file, cache_dir)
                    )
                    os.makedirs(os.path.dirname(target_file), exist_ok=True)
                    shutil.move(source_file, target_file)
            from tooldelta.plugin_manager import plugin_manager
            # 注册插件
            plugin_manager.push_plugin_reg_data(plugin_data)   
            Print.print_suc(f"成功下载插件 §f{plugin_data.name}§a 至插件文件夹")
        finally:
            # Clean up cache directory
            shutil.rmtree(cache_dir)

    def find_dirs(self, plugin_data: PluginRegData):
        try:
            data = _get_json_from_url(
                _url_join(self.plugins_download_url, "directory.json")
            )
            data_list = []
            for folder, files in data.items():
                if plugin_data.name == folder.split("/")[0]:
                    # 展开
                    for file in files:
                        data_list.append(folder + r"/" + file)
            return data_list
        except KeyError as err:
            Print.print_err(f"获取插件市场插件目录结构出现问题: 无法找到 {err}, 有可能是未来得及更新目录")
            return
        except Exception as err:
            Print.print_err(f"获取插件市场插件目录结构出现问题: {err}")
            return

    @staticmethod
    def get_latest_plugin_version(plugin_type: str, plugin_name: str):
        try:
            src_url = Cfg().get_cfg("ToolDelta基本配置.json", {"插件市场源": str})["插件市场源"]
        except:
            src_url = "https://mirror.ghproxy.com/raw.githubusercontent.com/ToolDelta/ToolDelta/main/plugin_market"
        return _get_json_from_url(
            _url_join(src_url, "latest_versions.json")
        )[plugin_type + "_plugin"].get(plugin_name)

market = PluginMarket()



from .basic_mods import *
from .packets import Packet_CommandOutput
from .start_tool_delta import start_tool_delta, safe_jump
from .start_tool_delta import plugins
from .plugin_load.PluginGroup import Plugin, PluginAPI
from .color_print import Print
from .frame import Config, Builtins, Frame, GameCtrl



import asyncio
from concurrent.futures import thread
import os
import sys
import importlib
import threading

from tooldelta.color_print import Print
from tooldelta.plugin_manager import plugin_manager
from tooldelta.plugin_load import plugin_is_enabled


# 定义插件处理函数列表
player_message_funcs = {}
player_join_funcs = {}
player_left_funcs = {}
player_death_funcs = {}
repeat_funcs = {}
init_plugin_funcs = {}


def player_message(priority=None):
    def decorator(func):
        player_message_funcs[func] = priority
        return func

    return decorator


def player_join(priority=None):
    def decorator(func):
        player_join_funcs[func] = priority
        return func

    return decorator


def player_left(priority=None):
    def decorator(func):
        player_left_funcs[func] = priority
        return func

    return decorator


def player_death(priority=None):
    def decorator(func):
        player_death_funcs[func] = priority
        return func

    return decorator


def init(priority=None):
    def decorator(func):
        init_plugin_funcs[func] = priority
        return func

    return decorator


def repeat(*args):
    def decorator(func):
        repeat_funcs[func] = args[0]
        return func

    return decorator


# repeat_task
async def repeat_task(func, time):
    while True:
        await asyncio.sleep(time)
        # 防止出错
        try:
            await func()
        except Exception as e:
            Print.print_err(f"repeat_task error: {e}")


async def execute_asyncio_task(func_dict: dict, *args, **kwargs):
    tasks = []
    none_tasks = []

    # 将任务添加到 tasks 列表或 none_tasks 列表中
    for func, priority in func_dict.items():
        if priority is not None:
            tasks.append((priority, func(*args, **kwargs)))
        else:
            none_tasks.append((priority, func(*args, **kwargs)))

    # 按优先级对非 None 任务排序
    tasks.sort(key=lambda x: x[0])

    # 将 none_tasks 列表附加到已排序的任务列表的末尾
    tasks += none_tasks

    await asyncio.gather(*[task[1] for task in tasks])


# 并发初始化插件
async def execute_init():
    await execute_asyncio_task(init_plugin_funcs)


async def run_repeat():
    # 为字典中的每一个函数创建一个循环特定时间的任务
    tasks = []
    for func, time in repeat_funcs.items():
        tasks.append(repeat_task(func, time))
    # 并发执行所有任务
    await asyncio.gather(*tasks)


async def safe_jump():
    try:
        main_task.cancel()
    except NameError:
        return


main_task: asyncio.Task


async def execute_repeat():
    global main_task
    main_task = asyncio.create_task(run_repeat())
    try:
        await main_task
    except asyncio.CancelledError:
        Print.print_suc("重复任务 repeat_task 已退出！")


# 处理玩家消息并执行插件
async def execute_player_message(playername, message):
    await execute_asyncio_task(player_message_funcs, playername, message)


async def execute_death_message(playername, killer):
    await execute_asyncio_task(player_death_funcs, playername, killer)


async def execute_player_join(playername):
    await execute_asyncio_task(player_join_funcs, playername)


async def execute_player_left(playername):
    await execute_asyncio_task(player_left_funcs, playername)


async def load_plugin_file(file):
    # 导入插件模块
    module_name = file
    sys.path.append(os.path.join(os.getcwd(), "插件文件", "ToolDelta注入式插件"))
    plugin_module = importlib.import_module(module_name)
    meta_data = create_plugin_metadata(
        getattr(plugin_module, "__plugin_meta__", {"name": module_name})
    )
    # 获取插件元数据
    plugin_manager.test_name_same(meta_data.name, file)
    if not plugin_manager.plugin_is_registered("injected", meta_data.name):
        plugin_manager.auto_register_plugin("injected", meta_data)
    return meta_data


class PluginMetadata:
    def __init__(
        self,
        name,
        author,
        description,
        version,
        usage,
        homepage,
    ):
        self.name = name
        self.author = author
        self.description = description
        self.usage = usage
        self.version = version
        self.homepage = homepage


def create_plugin_metadata(metadata_dict: dict):
    name = metadata_dict.get("name", "未命名插件")
    version = metadata_dict.get("version", "1.0")
    description = metadata_dict.get("description", "未知插件")
    usage = metadata_dict.get("usage", "")
    author = metadata_dict.get("author", "未知")
    homepage = metadata_dict.get("homepage", "")

    return PluginMetadata(name, author, description, version, usage, homepage)


async def load_plugin(plugin_grp):
    tasks = []

    # 读取本目录下的文件夹名字
    PLUGIN_PATH = os.path.join(os.getcwd(), "插件文件", "ToolDelta注入式插件")
    for file in os.listdir(PLUGIN_PATH):
        if not plugin_is_enabled(file):
            continue
        if os.path.isdir(os.path.join(PLUGIN_PATH, file)):
            plugin_grp.injected_plugin_loaded_num += 1
            task = asyncio.create_task(load_plugin_file(file))
            tasks.append(task)

    # 顺序加载插件并收集插件元数据
    all_plugin_metadata = []
    for task in tasks:
        plugin_metadata = await task
        all_plugin_metadata.append(plugin_metadata)

    # 打印所有插件的元数据
    for metadata in all_plugin_metadata:
        Print.print_suc(
            f"成功载入插件 {metadata.name} 版本: {metadata.version} 作者: {metadata.author}"
        )



import os
import re
import shutil
import socket
import platform
import time
from typing import Union
from concurrent.futures import ThreadPoolExecutor, as_completed
from tqdm import tqdm
import pyspeedtest
import requests

from .color_print import Print


def _pretty_kb(n: int) -> str:
    if n >= 1048576:
        return f"{round(n / 1048576, 2)}M"
    if n >= 1024:
        return f"{round(n / 1024, 2)}K"
    return f"{round(n, 1)}"


def _path_get_filename(path: str) -> Union[str, None]:
    if "/" not in path:
        return None
    return path.split("/")[-1]


def _is_common_text_file(url_path: str) -> bool:
    return any(
        url_path.endswith(i)
        for i in [".txt", ".yml", ".md", ".xml", ".py", ".h", ".c", ".pyi", ".json"]
    )


def get_file_size(url: str) -> Union[int, None]:
    response = requests.head(url, timeout=10)
    if "Content-Length" in response.headers:
        file_size = int(response.headers["Content-Length"])
        return file_size
    return None


def download_file_chunk(url: str, start_byte: int, end_byte: int, save_dir: str) -> int:
    headers = {"Range": f"bytes={start_byte}-{end_byte}"}
    response = requests.get(url, headers=headers, stream=True, timeout=10)
    response.raise_for_status()
    with open(save_dir + ".tmp", "rb+") as dwnf:
        dwnf.seek(start_byte)
        total_bytes = end_byte - start_byte + 1
        downloaded_bytes = 0
        for chunk in response.iter_content(chunk_size=8192):
            dwnf.write(chunk)
            downloaded_bytes += len(chunk)
        return downloaded_bytes


def download_file_singlethreaded(
    url: str, save_dir: str, ignore_warnings: bool = False
) -> None:
    with requests.get(url, stream=True, timeout=10) as res:
        res.raise_for_status()
        filesize = get_file_size(url)
        if filesize < 256 and not ignore_warnings:
            Print.print_war(f"下载 {url} 的文件警告: 文件大小异常, 不到 0.25KB")
        chunk_size = 8192
        nowsize = 0
        succ = False
        lastime = time.time()
        useSpeed = 0
        with open(save_dir + ".tmp", "wb") as dwnf:
            for chk in res.iter_content(chunk_size=8192):
                nowtime = time.time()
                if nowtime != lastime:
                    useSpeed = chunk_size / (nowtime - lastime)
                    lastime = nowtime
                nowsize += len(chk)
                dwnf.write(chk)
                if nowsize % 81920 == 0:  # 每下载 10 个数据块更新一次进度
                    prgs = nowsize / filesize
                    _tmp = int(prgs * 20)
                    bar = Print.colormode_replace(
                        "§f" + " " * _tmp + "§b" + " " * (20 - _tmp) + "§r ", 7
                    )
                    Print.print_with_info(
                        f"{bar} {_pretty_kb(nowsize)}B / {_pretty_kb(filesize)}B ({_pretty_kb(useSpeed)}B/s)    ",
                        "§a 下载 §r",
                        end="\r",
                        need_log=False,
                    )
        succ = True
        if succ:
            shutil.move(save_dir + ".tmp", save_dir)
        else:
            os.remove(save_dir + ".tmp")


def download_file_multithreading(
    url: str, save_dir: str, num_threads: int = 8, ignore_warnings: bool = False
) -> None:
    filesize = get_file_size(url)
    if filesize is None:
        download_file_singlethreaded(url=url, save_dir=save_dir)
    if filesize < 256 and not ignore_warnings:
        Print.print_war(f"下载 {url} 的文件警告: 文件大小异常, 不到 0.25KB")
    elif filesize < 1048576 and not ignore_warnings:
        download_file_singlethreaded(url=url, save_dir=save_dir)
    chunk_size = filesize // num_threads  # 每个线程下载的块大小
    with open(save_dir + ".tmp", "wb") as dwnf:
        with tqdm(
            total=filesize,
            unit="B",
            unit_scale=True,
            desc=Print.fmt_info("", "§a 下载 §r"),
            ncols=80,
        ) as pbar:

            def update_progress(downloaded_bytes: int) -> None:
                pbar.update(downloaded_bytes)

            with ThreadPoolExecutor(max_workers=num_threads) as executor:
                futures = []
                for i in range(num_threads):
                    start_byte = i * chunk_size
                    end_byte = start_byte + chunk_size - 1
                    if i == num_threads - 1:
                        end_byte = filesize - 1
                    future = executor.submit(
                        download_file_chunk, url, start_byte, end_byte, save_dir
                    )
                    future.add_done_callback(lambda f: update_progress(f.result()))
                    futures.append(future)
                for future in futures:
                    future.result()
    shutil.move(save_dir + ".tmp", save_dir)


def download_unknown_file(url: str, save_dir: str) -> None:
    # 鉴于 Content-Length 不一定表示文件原始大小, 二进制文件与文本文件需要分开下载
    resp = requests.get(url, timeout=10)
    resp.raise_for_status()

    with open(save_dir, "wb") as f:
        f.write(resp.content)


def test_site_latency(Da: dict) -> list:
    tmp_speed = {}
    urls = [Da["url"]] + Da["mirror_url"]
    with ThreadPoolExecutor() as executor:
        futures = [executor.submit(measure_latencyt, url) for url in urls]
        for future, url in zip(as_completed(futures), urls):
            latency = future.result()
            if latency != -1:
                tmp_speed[url] = latency

    sorted_speed = sorted(tmp_speed.items(), key=lambda x: x[1], reverse=True)
    return sorted_speed


def measure_latencyt(url: str) -> Union[float, int]:
    try:
        st = pyspeedtest.SpeedTest(
            re.search(r"(?<=http[s]://)[.\w-]*(:\d{,8})?((?=/)|(?!/))", url).group()
        )
        download_speed = st.download()  # / 1000000  # # 转换为兆字节/秒 （取消）
        return download_speed
    except Exception:
        return -1  # 返回-1表示测速失败


def get_free_port(start: int = 8080, end: int = 65535) -> int:
    for port in range(start, end):
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            try:
                s.bind(("localhost", port))
                return port
            except OSError:
                Print.print_war(f"端口 {port} 正被占用, 跳过")
    raise Exception(f"未找到空闲端口({start}~{end})")



from tooldelta.color_print import Print
from tooldelta.start_tool_delta import start_tool_delta
from tooldelta.plugin_manager import plugin_manager
from tooldelta.plugin_market import market
from tooldelta.sys_args import sys_args_to_dict

# TODO: 这是启动界面, 在此方法下写启动选项(启动ToolDelta, 插件管理, 插件市场)
def client_title():
    launch_mode = sys_args_to_dict()
    if launch_mode.get("l"):
        r = launch_mode["l"]
    else:
        Print.clean_print("§b请选择启动模式(使用启动参数 -l <启动模式> 可以跳过该页面):")
        Print.clean_print("1 - 启动 ToolDelta")
        Print.clean_print("2 - 打开 ToolDelta 插件管理器")
        Print.clean_print("3 - 打开 ToolDelta 插件市场")
        r = input("请选择:").strip()
    match r:
        case "1":
            start_tool_delta()
        case "2":
            plugin_manager.manage_plugins()
        case "3":
            market.enter_plugin_market()
        case _:
            Print.clean_print("§c不合法的模式: " + r)


from .basic_mods import os, requests, re
from .color_print import Print

class Game_Texts(object):
    def __init__(self) -> None:
        # 检测更新!
        # if not os.path.exists("data/game_texts"):
        #     version_match = re.match(r"v(\d+\.\d+\.\d+)", requests.get("https://api.github.com/repos/ToolDelta/ToolDelta-Game_Texts/releases/latest").json()["tag_name"])
        #     open("data/game_texts/version","w").write(version_match)
        #     packets_url = f"https://github.com/ToolDelta/ToolDelta-Game_Texts/releases/download/{version_match}"
        pass


from subprocess import Popen, PIPE
from .color_print import Print
from os import sep
from sys import argv

neccessary_libs = {
    "requests": None,
    "psutil": None,
    "nbt": None,
    "flask": None,
    "rich": None,
    "pymysql": None,
    "websockets": None,
    "qrcode": None,
    "ujson": None,
    "pyspeedtest": None,
    "tqdm": None,
}


def try_install_libs(parent_err):
    if argv[-1].split(sep)[-1] == "main.py":
        install_libs(neccessary_libs.keys())
    else:
        raise ModuleNotFoundError(f"打包程序内模块未能找到: {parent_err}")


def check_pip():
    try:
        p = Popen(
            [
                "python",
                "-m",
                "pip",
                "config",
                "set",
                "global.index-url",
                "https://pypi.tuna.tsinghua.edu.cn/simple",
            ],
            # shell=True,
            stdout=PIPE,
            stderr=PIPE,
        )
        p.wait()
    except Exception as err:
        Print.print_err(err)
        return False
    return not p.returncode


def get_single_lib(lib_name, lib_show_name=""):
    lib_show_name = lib_show_name or lib_name
    cmd = f"pip install {lib_name}"
    Print.print_inf(f"正在安装库: {lib_show_name}", end="\r")
    pipe = Popen(cmd.split(), stdout=PIPE, stderr=PIPE)
    pipe.wait()
    if pipe.returncode:
        Print.print_err(f"未成功安装库: {lib_show_name}")
        raise SystemExit
    Print.print_suc(f"成功安装库: {lib_show_name}")


def install_libs(libs):
    if not check_pip():
        Print.print_err("pip 环境未就绪!")
        raise SystemExit
    for lib in libs:
        get_single_lib(lib)



import ctypes

GoInt = ctypes.c_longlong
GoString = ctypes.c_char_p
GoBytes = ctypes.POINTER(ctypes.c_char)


class intGoSlice(ctypes.Structure):
    _fields_ = [
        ("data", ctypes.POINTER(ctypes.c_longlong)),
        ("len", ctypes.c_longlong),
        ("cap", ctypes.c_longlong),
    ]


class byteGoSlice(ctypes.Structure):
    _fields_ = [
        ("data", ctypes.POINTER(ctypes.c_char)),
        ("len", ctypes.c_longlong),
        ("cap", ctypes.c_longlong),
    ]


class ConnectFB_return(ctypes.Structure):
    _fields_ = [("connID", GoInt), ("err", GoString)]


class RecvGamePacket_return(ctypes.Structure):
    _fields_ = [("pktBytes", GoBytes), ("l", GoInt), ("err", GoString)]


class SendWSCommand_return(ctypes.Structure):
    _fields_ = [("uuid", GoString), ("err", GoString)]


class SendMCCommand_return(ctypes.Structure):
    _fields_ = [("uuid", GoString), ("err", GoString)]


class GamePacketBytesAsIsJsonStr_return(ctypes.Structure):
    _fields_ = [("jsonStr", GoString), ("err", GoString)]


class JsonStrAsIsGamePacketBytes_return(ctypes.Structure):
    _fields_ = [("pktBytes", GoBytes), ("l", GoInt), ("err", GoString)]


LIB: ctypes.CDLL


def InitLib(LIBs):
    global LIB
    LIB = LIBs
    # struct ConnectFB_return ConnectFB(char* address);
    LIB.ConnectFB.argtypes = [GoString]
    LIB.ConnectFB.restype = ConnectFB_return

    # ReleaseConnByID(GoInt id);
    LIB.ReleaseConnByID.argtypes = [GoInt]

    # struct RecvGamePacket_return RecvGamePacket(GoInt connID);
    LIB.RecvGamePacket.argtypes = [GoInt]
    LIB.RecvGamePacket.restype = RecvGamePacket_return

    # char* SendGamePacketBytes(GoInt connID, GoSlice content);
    LIB.SendGamePacketBytes.argtypes = [GoInt, byteGoSlice]
    LIB.SendGamePacketBytes.restype = GoString

    # char* SendFBCommand(GoInt connID, char* cmd);
    LIB.SendFBCommand.argtypes = [GoInt, GoString]
    LIB.SendFBCommand.restype = GoString

    # struct SendWSCommand_return SendWSCommand(GoInt connID, char* cmd);
    LIB.SendWSCommand.argtypes = [GoInt, GoString]
    LIB.SendWSCommand.restype = SendWSCommand_return

    # struct SendMCCommand_return SendMCCommand(GoInt connID, char* cmd);
    LIB.SendMCCommand.argtypes = [GoInt, GoString]
    LIB.SendMCCommand.restype = SendMCCommand_return

    # struct SendNoResponseCommand(GoInt connID, char* cmd);
    LIB.SendNoResponseCommand.argtypes = [GoInt, GoString]
    LIB.SendNoResponseCommand.restype = GoString

    # struct GamePacketBytesAsIsJsonStr_return GamePacketBytesAsIsJsonStr(char* pktBytes);
    LIB.GamePacketBytesAsIsJsonStr.argtypes = [byteGoSlice]
    LIB.GamePacketBytesAsIsJsonStr.restype = GamePacketBytesAsIsJsonStr_return

    # struct JsonStrAsIsGamePacketBytes_return JsonStrAsIsGamePacketBytes(GoInt packetID, char* jsonStr);
    LIB.JsonStrAsIsGamePacketBytes.argtypes = [GoInt, GoString]
    LIB.JsonStrAsIsGamePacketBytes.restype = JsonStrAsIsGamePacketBytes_return

    # char* CreatePacketInJsonStrByID(GoInt packetID);
    # LIB.CreatePacketInJsonStrByID.argtypes = [GoInt]
    # LIB.CreatePacketInJsonStrByID.restype = GoString

    LIB.FreeMem.argtypes = [ctypes.c_void_p]
    return LIB


def to_GoInt(i: int):
    return ctypes.c_longlong(i)


def to_PyInt(i):
    return i


def to_GoString(string: str):
    return ctypes.c_char_p(bytes(string, encoding="utf-8"))


def to_PyString(string: bytes):
    return string.decode(encoding="utf-8")


def to_GoByteSlice(bs: bytes):
    l = len(bs)
    kwargs = {
        "data": (ctypes.c_char * l)(*bs),
        "len": l,
        "cap": l,
    }
    return byteGoSlice(**kwargs)


def freeMem(address):
    LIB.FreeMem(address)


def check_err_in_struct(r):
    if r.err is not None:
        err = to_PyString(r.err)
        raise Exception(err)


def check_err(r):
    if r is not None:
        err = to_PyString(r)
        raise Exception(err)


def ConnectFB(address: str) -> int:
    r = LIB.ConnectFB(to_GoString(address))
    check_err_in_struct(r)
    return r.connID


def ReleaseConnByID(connID: int) -> None:
    LIB.ReleaseConnByID(to_GoInt(connID))


def RecvGamePacket(connID: int):
    try:
        while 1:
            r = LIB.RecvGamePacket(to_GoInt(connID))
            check_err_in_struct(r)
            bs: bytes = r.pktBytes[: r.l]
            freeMem(r.pktBytes)
            yield bs
    except:
        return


def RecvGamePacketIt(connID: int):
    while 1:
        r = LIB.RecvGamePacket(to_GoInt(connID))
        check_err_in_struct(r)
        bs = r.pktBytes[: r.l]
        freeMem(r.pktBytes)
        yield bs


# not tested
def SendGamePacketBytes(connID: int, content: bytes) -> None:
    inp = to_GoByteSlice(content)
    r = LIB.SendGamePacketBytes(connID, inp)
    check_err(r)


def SendFBCommand(connID: int, cmd: str) -> None:
    r = LIB.SendFBCommand(to_GoInt(connID), to_GoString(cmd))
    check_err(r)


def SendNoResponseCommand(connID: int, cmd: str) -> None:
    r = LIB.SendNoResponseCommand(to_GoInt(connID), to_GoString(cmd))
    check_err(r)


def SendMCCommand(connID: int, cmd: str) -> str:
    r = LIB.SendMCCommand(to_GoInt(connID), to_GoString(cmd))
    check_err_in_struct(r)
    uuid = r.uuid[:]
    return uuid


def SendWSCommand(connID: int, cmd: str) -> str:
    r = LIB.SendWSCommand(to_GoInt(connID), to_GoString(cmd))
    check_err_in_struct(r)
    uuid = r.uuid[:]
    return uuid


def GamePacketBytesAsIsJsonStr(pktBytes: bytes) -> str:
    r = LIB.GamePacketBytesAsIsJsonStr(to_GoByteSlice(pktBytes))
    check_err_in_struct(r)
    jsonStr = to_PyString(r.jsonStr)
    return jsonStr


def JsonStrAsIsGamePacketBytes(packetID: int, jsonStr: str) -> bytes:
    r = LIB.JsonStrAsIsGamePacketBytes(to_GoInt(packetID), to_GoString(jsonStr))
    check_err_in_struct(r)
    bs = r.pktBytes[: r.l]
    freeMem(r.pktBytes)
    return bs


def inspectPacketID(packet: bytes):
    return packet[0]



import datetime
import colorama

# 防止输出颜色代码
colorama.init(autoreset=True)

try:
    from .logger import publicLogger
except:
    pass

unicode = type("")


def simple_fmt(kw, __sub):
    for k, v in kw.items():
        if k in __sub:
            __sub = __sub.replace(k, str(v))
    return __sub


class _Print:
    INFO_NORMAL = "§f 信息 "
    INFO_WARN = "§6 警告 "
    INFO_ERROR = "§4 报错 "
    INFO_FAIL = "§c 失败 "
    INFO_SUCC = "§a 成功 "
    INFO_LOAD = "§d 加载 "
    STD_COLOR_LIST = [
        ["0", "#000000"],
        ["1", "#0000AA"],
        ["2", "#00AA00"],
        ["3", "#00AAAA"],
        ["4", "#AA0000"],
        ["5", "#AA00AA"],
        ["6", "#FFAA00"],
        ["7", "#AAAAAA"],
        ["8", "#555555"],
        ["9", "#5555FF"],
        ["a", "#55FF55"],
        ["b", "#55FFFF"],
        ["c", "#FF5555"],
        ["d", "#FF55FF"],
        ["e", "#FFFF55"],
        ["f", "#FFFFFF"],
        ["g", "#DDD605"],
        ["r", "/"],
    ]

    def __call__(self, text: str):
        self.print_inf(text)

    def colormode_replace(self, text: str, showmode=0):
        # 1 = bg_color
        text = self._strike(text)
        return (
            simple_fmt(
                {
                    "§1": f"\033[{showmode};37;34m",
                    "§2": f"\033[{showmode};37;32m",
                    "§3": f"\033[{showmode};37;36m",
                    "§4": f"\033[{showmode};37;31m",
                    "§5": f"\033[{showmode};37;35m",
                    "§6": f"\033[{showmode};37;33m",
                    "§7": f"\033[{showmode};37;90m",
                    "§8": f"\033[{showmode};37;2m",
                    "§9": f"\033[{showmode};37;94m",
                    "§a": f"\033[{showmode};37;92m",
                    "§b": f"\033[{showmode};37;96m",
                    "§c": f"\033[{showmode};37;91m",
                    "§d": f"\033[{showmode};37;95m",
                    "§e": f"\033[{showmode};37;93m",
                    "§f": f"\033[{showmode};37;1m",
                    "§r": "\033[0m",
                    "§u": "\033[4m",
                    "§l": "\033[1m",
                },
                text,
            )
            + "\033[0m"
        )

    @staticmethod
    def align(text: str, length: int):
        l = len(text)
        for char in text:
            if not char.isascii():
                l += 1
        return text + " " * (length - l)

    @staticmethod
    def _strike(text: str):
        text_ok = ""
        strikeMode = False
        i = 0
        while i < len(text):
            char = text[i]
            try:
                if char == "§":
                    if text[i + 1] == "s":
                        strikeMode = True
                        i += 2
                        continue
                    if text[i + 1] == "r":
                        strikeMode = False
            except IndexError:
                pass
            if strikeMode:
                text_ok += "\u0336" + char
            else:
                text_ok += char
            i += 1
        return text_ok

    def print_with_info(
        self, text: str, info: str = INFO_NORMAL, need_log=True, **print_kwargs
    ):
        if need_log:
            self.c_log(info, text)
        setNextColor = "§r"
        if "\n" in text:
            output_txts = []
            for text_line in str(text).split("\n"):
                if "§" in text_line:
                    try:
                        n = text_line.rfind("§")
                        _setNextCol = text_line[n : n + 2]
                        if setNextColor == -1:
                            raise AssertionError
                        setNextColor = _setNextCol
                    except:
                        pass
                output_txts.append(
                    datetime.datetime.now().strftime("[%H:%M] ")
                    + self.colormode_replace(info, 7)
                    + " "
                    + self.colormode_replace(setNextColor + text_line)
                )
            print("\n".join(output_txts), **print_kwargs)
        else:
            print(
                datetime.datetime.now().strftime("[%H:%M] ")
                + self.colormode_replace(info, 7)
                + " "
                + self.colormode_replace(text),
                **print_kwargs,
            )

    @staticmethod
    def align(text: str, length=15):
        l = len(text)
        for i in text:
            if not i.isascii():
                l += 1
        return text + " " * (length - l)

    def clean_print(self, text: str, **print_kwargs):
        print(self.colormode_replace(text), **print_kwargs)

    def clean_fmt(self, text: str):
        return self.colormode_replace(text)

    def print_err(self, text: str, **print_kwargs):
        self.print_with_info(f"§c{text}", self.INFO_ERROR, **print_kwargs)

    def print_inf(self, text: str, **print_kwargs):
        self.print_with_info(f"{text}", self.INFO_NORMAL, **print_kwargs)

    def print_suc(self, text: str, **print_kwargs):
        self.print_with_info(f"§a{text}", self.INFO_SUCC, **print_kwargs)

    def print_war(self, text: str, **print_kwargs):
        self.print_with_info(f"§6{text}", self.INFO_WARN, **print_kwargs)

    def print_load(self, text: str, **print_kwargs):
        self.print_with_info(f"§d{text}", self.INFO_LOAD, **print_kwargs)

    def fmt_info(self, text: str, info: str = "§f 信息 "):
        setNextColor = "§r"
        if "\n" in text:
            output_txts = []
            for text_line in str(text).split("\n"):
                if "§" in text_line:
                    try:
                        n = text_line.rfind("§")
                        _setNextCol = text_line[n : n + 2]
                        if setNextColor == -1:
                            raise AssertionError
                        setNextColor = _setNextCol
                    except:
                        pass
                output_txts.append(
                    datetime.datetime.now().strftime("[%H:%M] ")
                    + self.colormode_replace(info, 7)
                    + " "
                    + self.colormode_replace(setNextColor + text_line)
                )
            return "\n".join(output_txts)
        return (
            datetime.datetime.now().strftime("[%H:%M] ")
            + self.colormode_replace(info, 7)
            + " "
            + self.colormode_replace(text)
        )

    def c_log(self, inf: str, msg: str):
        for _g, _s in [
            ("§6 警告 ", "WARN"),
            ("§a 成功 ", "INFO"),
            ("§f 信息 ", "INFO"),
            ("§c 失败 ", "FAIL"),
            ("§4 报错 ", "ERROR"),
        ]:
            if inf == _g:
                inf = _s
                break
        for col, _ in self.STD_COLOR_LIST:
            col = "§" + col
            msg = msg.replace(col, "")
        for col, _ in self.STD_COLOR_LIST:
            col = "§" + col
            inf = inf.replace(col, "")
        inf = inf.replace(" ", "")
        publicLogger.log_in(msg, inf)


Print = _Print()



NON_FUNC = lambda *_: None


class PluginSkip(EOFError):
    ...

class PluginRegData:
    def __init__(self, name: str, plugin_data: dict = None, is_registered=True, is_enabled=True):
        if plugin_data is None:
            plugin_data = {}
        self.name: str = name
        if isinstance(plugin_data.get("version"), str):
            self.version: tuple = tuple(
                int(i) for i in plugin_data.get("version", "0.0.0").split(".")
            )
        else:
            self.version = plugin_data.get("version", (0, 0, 0))
        self.author: str = plugin_data.get("author", "unknown")
        self.plugin_type: str = plugin_data.get("plugin-type", "unknown")
        self.description: str = plugin_data.get("description", "")
        self.pre_plugins: dict[str, str] = plugin_data.get("pre-plugins", {})
        self.is_registered = is_registered
        if plugin_data.get("enabled") is not None:
            self.is_enabled = plugin_data.get("enabled")
        else:
            self.is_enabled = is_enabled

    def dump(self):
        return {
            "author": self.author,
            "version": ".".join([str(i) for i in self.version]),
            "plugin-type": self.plugin_type,
            "description": self.description,
            "pre-plugins": self.pre_plugins,
            "enabled": self.is_enabled
        }

    @property
    def version_str(self):
        return ".".join(str(i) for i in self.version)

    @property
    def plugin_type_str(self):
        return {
            "classic": "组合式",
            "injected": "注入式",
            "dotcs": "DotCS",
            "unknown": "未知类型",
        }.get(self.plugin_type, "未知类型")

def plugin_is_enabled(pname: str):
    return not pname.endswith("+disabled")


# FURTURE library for Plugin Emulator
from .color_print import Print

FLOW_UP = 2**31 - 1
FLOW_DOWN = -(2**31)
FLOW_FIX = 2**32
scbs = {}


class ExecFailed(Exception):
    ...


def input_int(__prompt="", errmsg="请输入整数"):
    while 1:
        try:
            return int(input(Print.fmt_info(__prompt)))
        except ValueError:
            Print.print_err(errmsg)
            continue


def scb_exists(scbname):
    return scbname in scbs


def test_overflow(num):
    while num >= FLOW_UP:
        num -= FLOW_FIX
    while num <= FLOW_DOWN:
        num += FLOW_FIX


def require_score(scbname, tar_name, not_set_to_0=False):
    if not scb_exists(scbname):
        raise ExecFailed(f"计分板 {scbname} 不存在")
    if scbs[scbname].get(tar_name) is not None:
        return scbs[scbname][tar_name]
    if not_set_to_0:
        scbs[scbname][tar_name] = 0
        return 0
    score = input_int(f"需要计分板 {scbname} 的 {tar_name} 的分数, 请输入: ")
    scbs[scbname][tar_name] = score
    return score


def sendcmd(cmd: str, waitForResp=False, timeout=30):
    if cmd.startswith("/"):
        cmd = cmd[1:]
    cmds = cmd.split()
    try:
        match cmds[0]:
            case "scoreboard":
                if cmds[1] == "objectives":
                    if cmds[2] == "add":
                        scbname = cmds[3]
                        if cmds[4] != "dummy":
                            raise AssertionError
                        if scb_exists(scbname):
                            raise ExecFailed(f"计分板 {scbname} 已存在")
                        if len(scbname) not in range(16):
                            raise AssertionError("计分板名太长")
                        scbs[scbname] = {}
                    elif cmds[2] == "remove":
                        scbname = cmds[3]
                        if not scb_exists(scbname):
                            raise ExecFailed(f"计分板 {scbname} 不存在")
                    if cmds[2] != "setdisplay":
                        raise AssertionError
                elif cmds[1] == "players":
                    if cmds[2] in ["add", "remove"]:
                        # To be discussed
                        if cmd[2] == "remove":
                            fx = -1
                        else:
                            fx = 1
                        target, scbname, addscore = cmds[3:5]
                        if scbs.get(scbname) is None:
                            raise ExecFailed(f"计分板 {scbname} 不存在")
                        final = scbs[scbname].get(target, 0) + int(addscore) * fx
                        scbs[scbname][target] = test_overflow(final)
                    elif cmds[2] == "operation":
                        target1, scb1, op, target2, scb2 = cmds[3:8]
                        if op not in ["+=", "-=", "=", "*=", "/=", "%="]:
                            raise AssertionError("计分板运算符不合法")
                        match op:
                            case "+=":
                                final = require_score(scb1, target1) + require_score(
                                    scb2, target2
                                )
                            case "-=":
                                final = require_score(scb1, target1) + require_score(
                                    scb2, target2
                                )
                            case "*=":
                                final = require_score(scb1, target1) * require_score(
                                    scb2, target2
                                )
                            case "/=":
                                final = int(
                                    require_score(scb1, target1)
                                    / require_score(scb2, target2)
                                )
                            case "%=":
                                final = require_score(scb1, target1) % require_score(
                                    scb2, target2
                                )
                        scbs[scbname][target1] = test_overflow(final)
                    elif cmds[2] == "reset":
                        target, scb = cmds[3:5]
                        require_score(scb, target)
                        del scbs[scbname][target]

    except (IndexError, AssertionError):
        ...
